import abc

from django.db import transaction
from typing import List, NoReturn

from ..constants import MIGRATION_TABLE_BACKUP_DIR_NAME, MIGRATION_TABLE_BACKUP_FILE_NAME, \
    MIGRATION_FILES_BACKUP_DIR_NAME
from ..utils import BackupFile, Migration, BackupDir, AppMigrationsDir


class BaseRestore(abc.ABC):

    @abc.abstractmethod
    def restore(self):
        raise NotImplementedError


class MigrationsTableRestore(BaseRestore):

    def __init__(self):
        backup_dir = BackupDir(MIGRATION_TABLE_BACKUP_DIR_NAME)
        self.file_handler = BackupFile(
            directory=backup_dir,
            file_name=MIGRATION_TABLE_BACKUP_FILE_NAME
        )

    @transaction.atomic
    def restore(self) -> NoReturn:
        backup_migrations = self.get_migrations_data_from_backup()
        Migration.objects.all().delete()
        Migration.objects.bulk_create(backup_migrations)

    def get_migrations_data_from_backup(self) -> List[Migration]:
        migrations_data = self.file_handler.read()
        backup_migrations = []
        for migration in migrations_data:
            backup_migrations.append(
                Migration(**migration)
            )

        return backup_migrations


class MigrationFilesRestore(BaseRestore):

    def __init__(self, app_name: str):
        self.app_name = app_name
        self.app_migrations_dir = AppMigrationsDir(app_name=app_name)
        self.migrations_backup_dir = BackupDir(MIGRATION_FILES_BACKUP_DIR_NAME, app_name)

    def restore(self) -> NoReturn:
        if not self.app_migrations_dir.has_migration:
            return

        self.migrations_backup_dir.copy(destination=self.app_migrations_dir.path)
        self.migrations_backup_dir.clear()



from datetime import datetime
from pathlib import Path
from shutil import rmtree
from unittest.mock import patch

from django.db.migrations.loader import MIGRATIONS_MODULE_NAME
from django.db.migrations.recorder import MigrationRecorder
from django.test import TestCase
from django.utils.timezone import now
from django.conf import settings

from zero_migrations.apps import ZeroMigrationsConfig
from zero_migrations.exceptions import BackupError
from zero_migrations.utils import BackupDir
from zero_migrations.utils.backup import BackupFile, MigrationsTableBackup, AppMigrationsDir
from zero_migrations.utils.restore import MigrationsTableRestore


class TestBackupDir(TestCase):

    def setUp(self) -> None:
        self.test_backup_dir_name = "test"
        self.backup_dir = BackupDir(self.test_backup_dir_name)

    def test_path(self):
        self.assertEqual(
            str(self.backup_dir.path),
            str(Path(settings.BASE_DIR) / BackupDir.BACKUP_DIR_NAME / self.test_backup_dir_name)
        )

    @patch("zero_migrations.utils.os.listdir")
    def test_get_files_with_postfix(self, mock_listdir):
        mock_listdir.return_value = ["0001_sth.json", "0002_test.json"]

        files = self.backup_dir.get_files_with_postfix(postfix="test.json")
        self.assertEqual(files, ["0002_test.json"])


class TestAppMigrationsDir(TestCase):

    def setUp(self) -> None:
        self.test_app_name = ZeroMigrationsConfig.name
        self.migrations_dir = AppMigrationsDir(self.test_app_name)

    def test_path(self):
        self.assertEqual(
            str(self.migrations_dir.path),
            str(Path(settings.BASE_DIR) / MIGRATIONS_MODULE_NAME)
        )


class TestBackupFile(TestCase):

    def setUp(self) -> None:
        self.test_backup_dir_name = "test"

        self.backup_dir = BackupDir(self.test_backup_dir_name)
        self.backup_file = BackupFile(directory=self.backup_dir, file_name="test.json")

    @patch("zero_migrations.utils.os.listdir")
    def test_latest_revision_without_any_file_in_dir(self, mock_listdir):
        mock_listdir.return_value = []

        self.assertIsNone(self.backup_file.latest_revision)

    @patch("zero_migrations.utils.os.listdir")
    def test_latest_revision_with_files_in_dir(self, mock_listdir):
        mock_listdir.return_value = ["0001_test.json", "0002_test.json"]

        self.assertEqual(str(self.backup_file.latest_revision), "0002_test.json")

    @patch("zero_migrations.utils.os.listdir")
    def test_new_revision_without_any_file_in_dir(self, mock_listdir):
        mock_listdir.return_value = []

        self.assertEqual(str(self.backup_file.next_revision), f"{BackupFile.REVISION_START_FROM}_test.json")

    @patch("zero_migrations.utils.os.listdir")
    def test_new_revision_with_files_in_dir(self, mock_listdir):
        mock_listdir.return_value = ["0001_test.json", "0002_test.json"]

        self.assertEqual(str(self.backup_file.next_revision), "0003_test.json")

    @patch("zero_migrations.utils.os.listdir")
    def test_new_file_path_without_any_file_in_dir(self, mock_listdir):
        mock_listdir.return_value = []

        self.assertEqual(
            str(self.backup_file.new_file_path),
            str(self.backup_dir.path / f"{BackupFile.REVISION_START_FROM}_test.json")
        )

    @patch("zero_migrations.utils.os.listdir")
    def test_new_file_path_with_files_in_dir(self, mock_listdir):
        mock_listdir.return_value = ["0001_test.json", "0002_test.json"]

        self.assertEqual(
            str(self.backup_file.new_file_path),
            str(self.backup_dir.path / "0003_test.json")
        )

    def test_write_data(self):
        migrations_data = [
            {
                "id": 1,
                "app": "test",
                "name": "test",
                "applied": now()
            },
            {
                "id": 2,
                "app": "test1",
                "name": "test1",
                "applied": now()
            }
        ]
        self.backup_file.write(data=migrations_data)

        written_data = self.backup_file.read()

        self.assertEqual(written_data[0]["id"], 1)
        self.assertEqual(written_data[0]["app"], "test")
        self.assertEqual(written_data[0]["name"], "test")
        self.assertTrue(isinstance(written_data[0]["applied"], datetime))

        self.assertEqual(written_data[1]["id"], 2)
        self.assertEqual(written_data[1]["app"], "test1")
        self.assertEqual(written_data[1]["name"], "test1")
        self.assertTrue(isinstance(written_data[1]["applied"], datetime))

    def tearDown(self) -> None:
        rmtree(self.backup_dir.path, ignore_errors=True)


class TestBackupMigrationsTable(TestCase):

    def setUp(self) -> None:
        self.backup_handler = MigrationsTableBackup()
        MigrationRecorder.Migration.objects.all().delete()

    def test_get_migrations_data_from_db(self):
        migration_1 = MigrationRecorder.Migration.objects.create(app="test1", name="test1")
        migration_2 = MigrationRecorder.Migration.objects.create(app="test2", name="test2")

        migrations = self.backup_handler.get_migrations_data_from_db()

        self.assertEqual(migrations[0]["id"], migration_1.id)
        self.assertEqual(migrations[0]["app"], migration_1.app)
        self.assertEqual(migrations[0]["name"], migration_1.name)
        self.assertEqual(migrations[0]["applied"], migration_1.applied)

        self.assertEqual(migrations[1]["id"], migration_2.id)
        self.assertEqual(migrations[1]["app"], migration_2.app)
        self.assertEqual(migrations[1]["name"], migration_2.name)
        self.assertEqual(migrations[1]["applied"], migration_2.applied)

    @patch("zero_migrations.utils.backup.MigrationsTableBackup.get_migrations_data_from_db")
    def test_backup_with_invalid_backup_raise_error(self, mock_get_migration_data):
        mock_get_migration_data.return_value = [
            {
                "id": 1,
                "app": "test",
                "name": "test",
                "applied": now()
            }
        ]

        MigrationRecorder.Migration.objects.create(app="test1", name="test1")
        MigrationRecorder.Migration.objects.create(app="test2", name="test2")

        with self.assertRaises(BackupError):
            self.backup_handler.backup()


class TestRestoreMigrationsTable(TestCase):

    def setUp(self) -> None:
        self.restore_handler = MigrationsTableRestore()

    @patch("zero_migrations.utils.BackupFile.read")
    def test_get_migrations_data_from_backup(self, mock_read):
        mock_read.return_value = [
            {
                "id": 1,
                "app": "test",
                "name": "test",
                "applied": now()
            },
            {
                "id": 2,
                "app": "test1",
                "name": "test1",
                "applied": now()
            }
        ]

        migrations = self.restore_handler.get_migrations_data_from_backup()

        self.assertEqual(migrations[0].id, 1)
        self.assertEqual(migrations[0].app, "test")
        self.assertEqual(migrations[0].name, "test")

        self.assertEqual(migrations[1].id, 2)
        self.assertEqual(migrations[1].app, "test1")
        self.assertEqual(migrations[1].name, "test1")

    @patch("zero_migrations.utils.BackupFile.read")
    def test_restore_data_from_backup_success(self, mock_read):
        mock_read.return_value = [
            {
                "id": 1,
                "app": "test",
                "name": "test",
                "applied": now()
            },
            {
                "id": 2,
                "app": "test1",
                "name": "test1",
                "applied": now()
            }
        ]

        self.restore_handler.restore()

        migrations = MigrationRecorder.Migration.objects.all()

        self.assertEqual(migrations.count(), 2)
        self.assertEqual(migrations[0].id, 1)
        self.assertEqual(migrations[0].app, "test")
        self.assertEqual(migrations[0].name, "test")

        self.assertEqual(migrations[1].id, 2)
        self.assertEqual(migrations[1].app, "test1")
        self.assertEqual(migrations[1].name, "test1")



import abc
from typing import List, NoReturn

from ..constants import MIGRATION_TABLE_BACKUP_FILE_NAME,\
    MIGRATION_TABLE_BACKUP_DIR_NAME,\
    MIGRATION_FILES_BACKUP_DIR_NAME
from ..exceptions import BackupError
from ..utils import BackupFile, Migration, BackupDir, AppMigrationsDir


class BaseBackup(abc.ABC):

    @abc.abstractmethod
    def backup(self):
        raise NotImplementedError


class MigrationsTableBackup(BaseBackup):

    def __init__(self):
        backup_directory = BackupDir(MIGRATION_TABLE_BACKUP_DIR_NAME)
        self.file_handler = BackupFile(
            directory=backup_directory,
            file_name=MIGRATION_TABLE_BACKUP_FILE_NAME
        )

    def backup(self) -> NoReturn:
        migrations_data = self.get_migrations_data_from_db()
        self._validate_backup()
        self.file_handler.write(data=migrations_data)

    @staticmethod
    def get_migrations_data_from_db() -> List[dict]:
        all_migrations = Migration.objects.iterator()
        data = []
        for migration in all_migrations:
            data.append(
                {field.name: getattr(migration, field.name, None)
                 for field in migration._meta.fields}
            )
        return data

    def _validate_backup(self) -> NoReturn:
        migrations_count_in_db = Migration.objects.count()
        migrations_count_in_file = len(self.get_migrations_data_from_db())

        if migrations_count_in_db != migrations_count_in_file:
            raise BackupError(f"There is an error in backup process. "
                              f"Apparently migrations count in db is: {migrations_count_in_db}, "
                              f"while its count in backup file is: {migrations_count_in_file}. "
                              f"We don't proceed any further, but you can take a backup yourself.")


class MigrationFilesBackup(BaseBackup):

    def __init__(self, app_name: str):
        self.app_name = app_name
        self.app_migrations_dir = AppMigrationsDir(app_name=app_name)
        self.migrations_backup_dir = BackupDir(MIGRATION_FILES_BACKUP_DIR_NAME, app_name)

    def backup(self) -> NoReturn:
        if not self.app_migrations_dir.has_migration:
            return

        self.app_migrations_dir.copy(destination=self.migrations_backup_dir.path)



import site
from typing import List, NoReturn
from functools import lru_cache

from django.apps import apps
from django.core.management import BaseCommand, call_command

from zero_migrations.utils import BackupDir, AppMigrationsDir
from zero_migrations.utils.backup import MigrationsTableBackup, MigrationFilesBackup
from zero_migrations.utils.restore import MigrationFilesRestore, MigrationsTableRestore


class Command(BaseCommand):
    help = "zeromigrations command to set migration files zero."

    MAKE_BACKUP = 1
    RESTORE_LAST_BACKUP = 2
    PROCEED = 3

    DELETE_MIGRATION_FILES = 1
    KEEP_MIGRATION_FILES = 2

    def handle(self, *args, **options):
        choice = int(input(
            self.style.WARNING(
                "I suggest to make a backups from both your "
                "migrations and django_migrations table (just in case).\n"
                f"1- make backup\n"
                f"2- restore last backup\n"
                f"3- just proceed\n"
            )
        ))
        if choice == self.MAKE_BACKUP:
            self.make_backup()
        if choice == self.RESTORE_LAST_BACKUP:
            self.restore()
        if choice == self.PROCEED:
            self.zero_migrations()

    def make_backup(self) -> NoReturn:
        """
        Making a backup for django_migrations table and migration files in each app.
        These backups will be used if any failure happened in zero_migrations process.
        """
        MigrationsTableBackup().backup()

        for app in self.get_apps():
            MigrationFilesBackup(app_name=app).backup()

        proceed_perm = input(
            self.style.SUCCESS(
                "Backup made successfully!.\n"
                f"Backups are in: {BackupDir().path}.\n"
                f"Shall I set migrations zero now? (y/n) "
            )
        )
        if proceed_perm == "y":
            self.zero_migrations()

    def zero_migrations(self):
        """
        Settings migrations zero.
        This process includes of 4 steps:
            1- Running `migrate --fake {app_name} zero` for each app
            2- Removing migrations files.
            3- Running `makemigrations` to make new initial migrations.
            4- Running `migrate --fake-initial` to fake made initial migrations.
        In case of any failure, it tries to restore the latest backup.
        """
        try:
            self.stdout.write(self.style.WARNING("Migrate zero each app:"))
            for app in self.get_apps():
                self.stdout.write(self.style.WARNING(f"App name: {app}"))
                call_command("migrate", "--fake", app, "zero", force_color=True)

            self.stdout.write(self.style.WARNING("Removing migrations:"))
            for app in self.get_apps():
                AppMigrationsDir(app_name=app).clear()
                self.stdout.write(self.style.SUCCESS(f"Removed migrations of {app}"))

            self.stdout.write(self.style.WARNING("Making migrations:"))
            call_command("makemigrations", force_color=True)

            self.stdout.write(self.style.WARNING("Migrate with fake initial:"))
            call_command("migrate", "--fake-initial", force_color=True)

        except Exception as err:
            self.stderr.write(
                self.style.ERROR(
                    f"Process failed because of {err}.\n"
                )
            )
            self.restore()

    def restore(self):
        """
        Restoring the latest backup, if any failure has happened or the user wants to do it itself.
        First django_migrations table would be restored and after that, migrations files would be restored.
        For migration files restoring, user has to decide whether his/her current migrations files
            should be deleted or not. If he/her decided not to delete his/her current migration files,
            then backup files would be replaced
        Note that backup data would be clear after restoring.
        """
        self.stdout.write(
            self.style.WARNING("Restoring ...")
        )
        MigrationsTableRestore().restore()
        self.stdout.write(
            self.style.SUCCESS("django_migrations table restored.")
        )
        choice = input(
            self.style.WARNING(
                "Trying to restore migration file.\n"
                f"1- Delete my current migration files and restore.\n"
                f"2- Keep my current migration files and replace restore.\n"
            )
        )
        delete_migrations = True if choice == self.DELETE_MIGRATION_FILES else False
        for app in self.get_apps():
            migration_files_restore = MigrationFilesRestore(app_name=app)
            if migration_files_restore.migrations_backup_dir.has_migration:
                if delete_migrations:
                    migration_files_restore.app_migrations_dir.clear()
                migration_files_restore.restore()
            else:
                self.stdout.write(
                    self.style.ERROR(
                        f"Couldn't find any backup for app {app} in"
                        f"{migration_files_restore.migrations_backup_dir.path}."
                    )
                )

    @lru_cache
    def get_apps(self) -> List[str]:
        """
        Get all user django apps.
        Apps that have been installed and are not written by user, would be excluded. This is because we don't want to
            set third-party packages migrations zero.
        :return: List of user apps names.
        """
        installed_app_path = site.getsitepackages()[0]
        return [
            app.name for app in apps.get_app_configs()
            if not str(app.path).startswith(str(installed_app_path))
        ]



import abc
import json
import os
from datetime import datetime, date
from functools import lru_cache
from pathlib import Path
from shutil import rmtree, copytree

from typing import NoReturn, List, Optional, Union

from django.apps import apps
from django.db.migrations.loader import MIGRATIONS_MODULE_NAME
from django.db.migrations.recorder import MigrationRecorder

from ..apps import ZeroMigrationsConfig

Migration = MigrationRecorder.Migration


class BaseDir(abc.ABC):
    """
    Base class for directories.
    Each subclass should implement path() property, as operations happen in that path.

    -----
    Base operation:
        create: For creating directory in returned path from `path` property.
        copy: To copy a directory to a destination. This method is used when we want to copy our files(like backups).
        get_files: Return all files in a directory.
    """

    def create(self) -> NoReturn:
        Path(self.path).mkdir(parents=True, exist_ok=True)

    def copy(self, destination: Union[Path, str]) -> NoReturn:
        copytree(self.path, destination, dirs_exist_ok=True)

    @property
    def has_migration(self) -> bool:
        """True if there is at least one file in the directory that doesn't start with '_' or '~'."""
        return any(file[0] not in '_~' for file in self.get_files())

    def get_files(self) -> List[str]:
        try:
            return os.listdir(self.path)
        except FileNotFoundError:
            return []

    @property
    @abc.abstractmethod
    def path(self):
        raise NotImplementedError


class BackupDir(BaseDir):
    """
    A class for backup directory.
    As backup dir name is set to "backups" by default, the path would be sth like: /home/.../zero_migrations/backups
    You can pass multiple directory names to its constructor, and they would be joined together in final path.

    -----
    Usage Examples:
        bkp_dir = BackupDir("test")
        bkp_dir.path    --->    /home/.../zero_migrations/backups/test

        bkp_dir = BackupDir("first", "second")
        bkp_dir.path    --->    /home/.../zero_migrations/backups/first/second
    """
    BACKUP_DIR_NAME = "backups"

    def __init__(self, *dir_names):
        self._dir_names = dir_names

    def clear(self) -> NoReturn:
        """
        Cleaning all files inside the directory.
        As the files have to be backup files, it's OK to remove all at once.
        This method is used when the backup is used, and we want to clear our old backups.
        """
        rmtree(self.path, ignore_errors=True)

    def get_files_with_postfix(self, postfix: str = "") -> List[str]:
        """
        Filtering files inside the directory, base on their postfix.
        This method is used when we want to get all files that have our postfix like `migration_backup.json`

        -----
        Usage Example:
            Assume self.get_files() return ["test.json", "test.txt"], then:
            self.get_files_with_postfix("json")     --->    ["test.json"]
        """
        return [
            str(dir_) for dir_ in self.get_files()
            if str(dir_).endswith(postfix)
        ]

    @property
    @lru_cache
    def path(self) -> Path:
        return self.app_dir_path / self.BACKUP_DIR_NAME / Path(*self._dir_names)

    @property
    def app_dir_path(self) -> Path:
        return Path(apps.get_app_config(ZeroMigrationsConfig.name).path)


class AppMigrationsDir(BaseDir):

    def __init__(self, app_name: str):
        self.app_name = app_name

    def clear(self) -> NoReturn:
        """
        Cleaning files inside the migrations' directory.
        As it is not our directory, and it is user app migrations directory, we just delete *.py to prevent deleting
            user's other files.
        This method is used when we have run `migrate --fake app zero` and we want to delete old migrations.
        """
        for file_name in self.get_files():
            if file_name != "__init__.py" and file_name.endswith(".py"):
                os.remove(self.path / file_name)

    @property
    @lru_cache
    def path(self) -> Path:
        return Path(apps.get_app_config(app_label=self.app_name).path) / MIGRATIONS_MODULE_NAME


class BackupFile:
    """
    This class is for handling `django_migrations` backup files issues.

    For now, we keep old backups of `django_migrations` too. For handling the duplicate issue,
        we use a number as a prefix for our backup files, called REVISION_NUMBER. This number increases based on last
        backup file in the directory, and if no files is in there, it is just REVISION_START_FROM.

    -----
    Usage Examples:
        Assume that we have a backup file called `0001_migration_backup.json` in backup directory, then:
        bkp_dir = BackupDir("migrations")
        bkp_file = BackupFile(directory=bkp_dir, file_name="test")
        bkp_file.latest_file_path   --->    /home/.../zero_migrations/backups/migrations/0001_migration_backup.json
        bkp_file.new_file_path   --->    /home/.../zero_migrations/backups/migrations/0002_migration_backup.json
        bkp_file.read()     ---> Reads from latest_file_path
        bkp_file.write(data)     ---> Writes in new_file_path
    """

    REVISION_START_FROM = "0001"

    def __init__(self, directory: BackupDir, file_name: str):
        self._file_name = file_name
        self.backup_dir = directory

        self._revision_num_len = len(self.REVISION_START_FROM)

    def write(self, data: List[dict]) -> NoReturn:
        """
        For writing in new_file_path (file with new revision).
        First, it makes sure the backup directory exists via invoking `create()` on self.backup_dir.
        Then, it invokes `json.dump` to get json of passed data.
        Note that because data has datetime in it (applied field in migration record) and `json.dump` is unable to
            serialize that, we change datetime objects to `iso format`.
        """
        self.backup_dir.create()

        # Changing datetime to iso format, so that json.dump can serialize it.
        def datetime_json_serialize(obj):
            if isinstance(obj, (datetime, date)):
                return obj.isoformat()

        with open(self.new_file_path, "w+") as f:
            json.dump(data, f, default=datetime_json_serialize)

    def read(self) -> List[dict]:
        """
        For reading from latest_file_path (file with the latest revision).
        It invokes `json.load` to get dict of json data.
        Note that because the json data has iso format in it, and we want datetime object instead of that,
            we change iso format to datetime object as a hook in `json.load`.
        """

        # Changing iso format to datetime object.
        def datetime_json_deserialize(obj: dict):
            for field, value in obj.items():
                try:
                    obj[field] = datetime.fromisoformat(value)
                except (ValueError, TypeError):
                    pass
            return obj

        with open(self.latest_file_path, "r") as f:
            return json.load(f, object_hook=datetime_json_deserialize)

    @property
    def new_file_path(self) -> Path:
        return self.backup_dir.path / self.next_revision

    @property
    def latest_file_path(self) -> Path:
        if self.latest_revision:
            return self.backup_dir.path / self.latest_revision

        return self.new_file_path

    @property
    def next_revision(self) -> str:
        """
        Getting next revision to write on.
        If the directory has no backup file already, then we don't have latest_revision. In this scenario,
            the next_revision is going to be first revision (self.REVISION_START_FROM).
        Otherwise, next_revision is the latest_revision + 1.
        """
        latest_revision = self.latest_revision
        if not latest_revision:
            return f"{self.REVISION_START_FROM}_{self._file_name}"

        next_revision_number = self.make_next_revision_number()
        return f"{next_revision_number}{latest_revision[self._revision_num_len:]}"

    @property
    def latest_revision(self) -> Optional[str]:
        all_backups = self.backup_dir.get_files_with_postfix(postfix=self._file_name)
        if not all_backups:
            return None

        # As our backup files have the same postfix, the last file name in the sorted list of file names,
        # is our latest_revision.
        return sorted(all_backups)[-1]

    def make_next_revision_number(self) -> str:
        new_revision_number = int(self.latest_revision[:self._revision_num_len]) + 1
        return "%0{revision_num_len}d".format(revision_num_len=self._revision_num_len) % (new_revision_number,)
